from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *


class EffectMode:
    """
    Class that represent an enumeration of values for EffectMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    none = 0
    frequency_modulation = 1
    sync_and_pulse_width = 2
    warp_and_fold = 3


class FilterRouting:
    """
    Class that represent an enumeration of values for FilterRouting
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    serial = 0
    parallel = 1
    split = 2


class ModulationSource:
    """
    Class that represent an enumeration of values for ModulationSource
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    amp_envelope = 0
    envelope_2 = 1
    envelope_3 = 2
    lfo_1 = 3
    lfo_2 = 4
    midi_velocity = 5
    midi_note = 6
    midi_pitch_bend = 7
    midi_channel_pressure = 8
    midi_mod_wheel = 9
    midi_random = 10


class UnisonMode:
    """
    Class that represent an enumeration of values for UnisonMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    none = 0
    classic = 1
    slow_shimmer = 2
    fast_shimmer = 3
    phase_sync = 4
    position_spread = 5
    random_note = 6


class VoiceCount:
    """
    Class that represent an enumeration of values for VoiceCount
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    two = 0
    three = 1
    four = 2
    five = 3
    six = 4
    seven = 5
    eight = 6


class Voicing:
    """
    Class that represent an enumeration of values for Voicing
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    mono = 0
    poly = 1


class WavetableDevice:
    """
    This class represents a Wavetable device.
    """

    class View:
        """
        Representing the view aspects of a device.
        """

        def add_is_collapsed_listener(self, listener: Callable) -> None:
            """
            add_is_collapsed_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "is_collapsed" has changed.

                C++ signature :
                    void add_is_collapsed_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def is_collapsed_has_listener(self, listener: Callable) -> bool:
            """
            is_collapsed_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "is_collapsed".

                C++ signature :
                    bool is_collapsed_has_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def remove_is_collapsed_listener(self, listener: Callable) -> None:
            """
            remove_is_collapsed_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "is_collapsed".

                C++ signature :
                    void remove_is_collapsed_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        @property
        def canonical_parent(self) -> Any:
            """
            Get the canonical parent of the View.
            """

        @property
        def is_collapsed(self) -> bool:
            """
            Get/Set/Listen if the device is shown collapsed in the device chain.
            """

        @is_collapsed.setter
        def is_collapsed(self, value: bool) -> None:
            pass

    def add_filter_routing_listener(self, listener: Callable) -> None:
        """
        add_filter_routing_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "filter_routing" has changed.

            C++ signature :
                void add_filter_routing_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_is_active_listener(self, listener: Callable) -> None:
        """
        add_is_active_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_active" has changed.

            C++ signature :
                void add_is_active_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_latency_in_ms_listener(self, listener: Callable) -> None:
        """
        add_latency_in_ms_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "latency_in_ms" has changed.

            C++ signature :
                void add_latency_in_ms_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_latency_in_samples_listener(self, listener: Callable) -> None:
        """
        add_latency_in_samples_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "latency_in_samples" has changed.

            C++ signature :
                void add_latency_in_samples_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_modulation_matrix_changed_listener(self, listener: Callable) -> None:
        """
        add_modulation_matrix_changed_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "modulation_matrix_changed" has changed.

            C++ signature :
                void add_modulation_matrix_changed_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_mono_poly_listener(self, listener: Callable) -> None:
        """
        add_mono_poly_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "mono_poly" has changed.

            C++ signature :
                void add_mono_poly_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_oscillator_1_effect_mode_listener(self, listener: Callable) -> None:
        """
        add_oscillator_1_effect_mode_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_1_effect_mode" has changed.

            C++ signature :
                void add_oscillator_1_effect_mode_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_1_wavetable_category_listener(self, listener: Callable) -> None:
        """
        add_oscillator_1_wavetable_category_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_1_wavetable_category" has changed.

            C++ signature :
                void add_oscillator_1_wavetable_category_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_1_wavetable_index_listener(self, listener: Callable) -> None:
        """
        add_oscillator_1_wavetable_index_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_1_wavetable_index" has changed.

            C++ signature :
                void add_oscillator_1_wavetable_index_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_1_wavetables_listener(self, listener: Callable) -> None:
        """
        add_oscillator_1_wavetables_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_1_wavetables" has changed.

            C++ signature :
                void add_oscillator_1_wavetables_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_2_effect_mode_listener(self, listener: Callable) -> None:
        """
        add_oscillator_2_effect_mode_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_2_effect_mode" has changed.

            C++ signature :
                void add_oscillator_2_effect_mode_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_2_wavetable_category_listener(self, listener: Callable) -> None:
        """
        add_oscillator_2_wavetable_category_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_2_wavetable_category" has changed.

            C++ signature :
                void add_oscillator_2_wavetable_category_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_2_wavetable_index_listener(self, listener: Callable) -> None:
        """
        add_oscillator_2_wavetable_index_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_2_wavetable_index" has changed.

            C++ signature :
                void add_oscillator_2_wavetable_index_listener(TWavetableDevicePyHandle,boost::python::api::object)

        add_oscillator_2_wavetable_index_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_2_wavetable_index" has changed.

            C++ signature :
                void add_oscillator_2_wavetable_index_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_oscillator_2_wavetables_listener(self, listener: Callable) -> None:
        """
        add_oscillator_2_wavetables_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "oscillator_2_wavetables" has changed.

            C++ signature :
                void add_oscillator_2_wavetables_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_parameter_to_modulation_matrix(self, parameter: object) -> int:
        """
        add_parameter_to_modulation_matrix( (WavetableDevice)self, (DeviceParameter)parameter) -> int :
            Add a non-pitch parameter to the modulation matrix.

            C++ signature :
                int add_parameter_to_modulation_matrix(TWavetableDevicePyHandle,TPyHandle<ATimeableValue>)
        """

    def add_parameters_listener(self, listener: Callable) -> None:
        """
        add_parameters_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "parameters" has changed.

            C++ signature :
                void add_parameters_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_poly_voices_listener(self, listener: Callable) -> None:
        """
        add_poly_voices_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "poly_voices" has changed.

            C++ signature :
                void add_poly_voices_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_unison_mode_listener(self, listener: Callable) -> None:
        """
        add_unison_mode_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "unison_mode" has changed.

            C++ signature :
                void add_unison_mode_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_unison_voice_count_listener(self, listener: Callable) -> None:
        """
        add_unison_voice_count_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "unison_voice_count" has changed.

            C++ signature :
                void add_unison_voice_count_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def add_visible_modulation_target_names_listener(self, listener: Callable) -> None:
        """
        add_visible_modulation_target_names_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "visible_modulation_target_names" has changed.

            C++ signature :
                void add_visible_modulation_target_names_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def filter_routing_has_listener(self, listener: Callable) -> bool:
        """
        filter_routing_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "filter_routing".

            C++ signature :
                bool filter_routing_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def get_modulation_target_parameter_name(self, target_index: int) -> str:
        """
        get_modulation_target_parameter_name( (WavetableDevice)self, (int)target_index) -> str :
            Get the parameter name of the modulation target at the given index.

            C++ signature :
                TString get_modulation_target_parameter_name(TWavetableDevicePyHandle,int)
        """

    def get_modulation_value(self, target_index: int, source: int) -> float:
        """
        get_modulation_value( (WavetableDevice)self, (int)target_index, (int)source) -> float :
            Get the value of a modulation amount for the given target-source connection.

            C++ signature :
                float get_modulation_value(TWavetableDevicePyHandle,int,int)
        """

    def is_active_has_listener(self, listener: Callable) -> bool:
        """
        is_active_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_active".

            C++ signature :
                bool is_active_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def is_parameter_modulatable(self, parameter: object) -> bool:
        """
        is_parameter_modulatable( (WavetableDevice)self, (DeviceParameter)parameter) -> bool :
            Indicate whether the parameter is modulatable. Note that pitch parameters only exist in python and must be handled there.

            C++ signature :
                bool is_parameter_modulatable(TWavetableDevicePyHandle,TPyHandle<ATimeableValue>)
        """

    def latency_in_ms_has_listener(self, listener: Callable) -> bool:
        """
        latency_in_ms_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "latency_in_ms".

            C++ signature :
                bool latency_in_ms_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def latency_in_samples_has_listener(self, listener: Callable) -> bool:
        """
        latency_in_samples_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "latency_in_samples".

            C++ signature :
                bool latency_in_samples_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def modulation_matrix_changed_has_listener(self, listener: Callable) -> bool:
        """
        modulation_matrix_changed_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "modulation_matrix_changed".

            C++ signature :
                bool modulation_matrix_changed_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def mono_poly_has_listener(self, listener: Callable) -> bool:
        """
        mono_poly_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "mono_poly".

            C++ signature :
                bool mono_poly_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def oscillator_1_effect_mode_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_1_effect_mode_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_1_effect_mode".

            C++ signature :
                bool oscillator_1_effect_mode_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_1_wavetable_category_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_1_wavetable_category_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_1_wavetable_category".

            C++ signature :
                bool oscillator_1_wavetable_category_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_1_wavetable_index_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_1_wavetable_index_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_1_wavetable_index".

            C++ signature :
                bool oscillator_1_wavetable_index_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_1_wavetables_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_1_wavetables_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_1_wavetables".

            C++ signature :
                bool oscillator_1_wavetables_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_2_effect_mode_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_2_effect_mode_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_2_effect_mode".

            C++ signature :
                bool oscillator_2_effect_mode_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_2_wavetable_category_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_2_wavetable_category_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_2_wavetable_category".

            C++ signature :
                bool oscillator_2_wavetable_category_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_2_wavetable_index_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_2_wavetable_index_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_2_wavetable_index".

            C++ signature :
                bool oscillator_2_wavetable_index_has_listener(TWavetableDevicePyHandle,boost::python::api::object)

        oscillator_2_wavetable_index_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_2_wavetable_index".

            C++ signature :
                bool oscillator_2_wavetable_index_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def oscillator_2_wavetables_has_listener(self, listener: Callable) -> bool:
        """
        oscillator_2_wavetables_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "oscillator_2_wavetables".

            C++ signature :
                bool oscillator_2_wavetables_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def parameters_has_listener(self, listener: Callable) -> bool:
        """
        parameters_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "parameters".

            C++ signature :
                bool parameters_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def poly_voices_has_listener(self, listener: Callable) -> bool:
        """
        poly_voices_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "poly_voices".

            C++ signature :
                bool poly_voices_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_filter_routing_listener(self, listener: Callable) -> None:
        """
        remove_filter_routing_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "filter_routing".

            C++ signature :
                void remove_filter_routing_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_is_active_listener(self, listener: Callable) -> None:
        """
        remove_is_active_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_active".

            C++ signature :
                void remove_is_active_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_latency_in_ms_listener(self, listener: Callable) -> None:
        """
        remove_latency_in_ms_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "latency_in_ms".

            C++ signature :
                void remove_latency_in_ms_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_latency_in_samples_listener(self, listener: Callable) -> None:
        """
        remove_latency_in_samples_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "latency_in_samples".

            C++ signature :
                void remove_latency_in_samples_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_modulation_matrix_changed_listener(self, listener: Callable) -> None:
        """
        remove_modulation_matrix_changed_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "modulation_matrix_changed".

            C++ signature :
                void remove_modulation_matrix_changed_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_mono_poly_listener(self, listener: Callable) -> None:
        """
        remove_mono_poly_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "mono_poly".

            C++ signature :
                void remove_mono_poly_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_oscillator_1_effect_mode_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_1_effect_mode_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_1_effect_mode".

            C++ signature :
                void remove_oscillator_1_effect_mode_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_1_wavetable_category_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_1_wavetable_category_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_1_wavetable_category".

            C++ signature :
                void remove_oscillator_1_wavetable_category_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_1_wavetable_index_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_1_wavetable_index_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_1_wavetable_index".

            C++ signature :
                void remove_oscillator_1_wavetable_index_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_1_wavetables_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_1_wavetables_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_1_wavetables".

            C++ signature :
                void remove_oscillator_1_wavetables_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_2_effect_mode_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_2_effect_mode_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_2_effect_mode".

            C++ signature :
                void remove_oscillator_2_effect_mode_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_2_wavetable_category_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_2_wavetable_category_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_2_wavetable_category".

            C++ signature :
                void remove_oscillator_2_wavetable_category_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_2_wavetable_index_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_2_wavetable_index_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_2_wavetable_index".

            C++ signature :
                void remove_oscillator_2_wavetable_index_listener(TWavetableDevicePyHandle,boost::python::api::object)

        remove_oscillator_2_wavetable_index_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_2_wavetable_index".

            C++ signature :
                void remove_oscillator_2_wavetable_index_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_oscillator_2_wavetables_listener(self, listener: Callable) -> None:
        """
        remove_oscillator_2_wavetables_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "oscillator_2_wavetables".

            C++ signature :
                void remove_oscillator_2_wavetables_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_parameters_listener(self, listener: Callable) -> None:
        """
        remove_parameters_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "parameters".

            C++ signature :
                void remove_parameters_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_poly_voices_listener(self, listener: Callable) -> None:
        """
        remove_poly_voices_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "poly_voices".

            C++ signature :
                void remove_poly_voices_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_unison_mode_listener(self, listener: Callable) -> None:
        """
        remove_unison_mode_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "unison_mode".

            C++ signature :
                void remove_unison_mode_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_unison_voice_count_listener(self, listener: Callable) -> None:
        """
        remove_unison_voice_count_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "unison_voice_count".

            C++ signature :
                void remove_unison_voice_count_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def remove_visible_modulation_target_names_listener(self, listener: Callable) -> None:
        """
        remove_visible_modulation_target_names_listener( (WavetableDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "visible_modulation_target_names".

            C++ signature :
                void remove_visible_modulation_target_names_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def set_modulation_value(self, target_index: int, source: int, value: float) -> None:
        """
        set_modulation_value( (WavetableDevice)self, (int)target_index, (int)source, (float)value) -> None :
            Set the value of a modulation amount for the given target-source connection.

            C++ signature :
                void set_modulation_value(TWavetableDevicePyHandle,int,int,float)
        """

    def store_chosen_bank(self, arg2: int, arg3: int) -> None:
        """
        store_chosen_bank( (Device)arg1, (int)arg2, (int)arg3) -> None :
            Set the selected bank in the device for persistency.

            C++ signature :
                void store_chosen_bank(TPyHandle<ADevice>,int,int)
        """

    def unison_mode_has_listener(self, listener: Callable) -> bool:
        """
        unison_mode_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "unison_mode".

            C++ signature :
                bool unison_mode_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def unison_voice_count_has_listener(self, listener: Callable) -> bool:
        """
        unison_voice_count_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "unison_voice_count".

            C++ signature :
                bool unison_voice_count_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    def visible_modulation_target_names_has_listener(self, listener: Callable) -> bool:
        """
        visible_modulation_target_names_has_listener( (WavetableDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "visible_modulation_target_names".

            C++ signature :
                bool visible_modulation_target_names_has_listener(TWavetableDevicePyHandle,boost::python::api::object)
        """

    @property
    def can_have_chains(self) -> bool:
        """
        Returns true if the device is a rack.
        """

    @property
    def can_have_drum_pads(self) -> bool:
        """
        Returns true if the device is a drum rack.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the Device.
        """

    @property
    def class_display_name(self) -> str:
        """
        Return const access to the name of the device's class name as displayed in Live's browser and device chain
        """

    @property
    def class_name(self) -> str:
        """
        Return const access to the name of the device's class.
        """

    @property
    def filter_routing(self) -> Any:
        """
        Return the current filter routing.
        """

    @property
    def is_active(self) -> bool:
        """
        Return const access to whether this device is active. This will be false bothwhen the device is off and when it's inside a rack device which is off.
        """

    @property
    def latency_in_ms(self) -> Any:
        """
        Returns the latency of the device in ms.
        """

    @property
    def latency_in_samples(self) -> Any:
        """
        Returns the latency of the device in samples.
        """

    @property
    def mono_poly(self) -> Any:
        """
        Return the current voicing mode.
        """

    @property
    def name(self) -> str:
        """
        Return access to the name of the device.
        """

    @property
    def oscillator_1_effect_mode(self) -> Any:
        """
        Return the current effect mode of the oscillator 1.
        """

    @property
    def oscillator_1_wavetable_category(self) -> Any:
        """
        Return the current wavetable category of the oscillator 1.
        """

    @property
    def oscillator_1_wavetable_index(self) -> Any:
        """
        Return the current wavetable index of the oscillator 1.
        """

    @property
    def oscillator_1_wavetables(self) -> Any:
        """
        Get a vector of oscillator 1's wavetable names.
        """

    @property
    def oscillator_2_effect_mode(self) -> Any:
        """
        Return the current effect mode of the oscillator 2.
        """

    @property
    def oscillator_2_wavetable_category(self) -> Any:
        """
        Return the current wavetable category of the oscillator 2.
        """

    @property
    def oscillator_2_wavetable_index(self) -> Any:
        """
        Return the current wavetable index of the oscillator 2.
        """

    @property
    def oscillator_2_wavetables(self) -> Any:
        """
        Get a vector of oscillator 2's wavetable names.
        """

    @property
    def oscillator_wavetable_categories(self) -> Any:
        """
        Get a vector of the available wavetable categories.
        """

    @property
    def parameters(self) -> list[DeviceParameter]:
        """
        Const access to the list of available automatable parameters for this device.
        """

    @property
    def poly_voices(self) -> Any:
        """
        Return the current number of polyphonic voices. Uses the VoiceCount enumeration.
        """

    @property
    def type(self) -> Any:
        """
        Return the type of the device.
        """

    @property
    def unison_mode(self) -> Any:
        """
        Return the current unison mode.
        """

    @property
    def unison_voice_count(self) -> Any:
        """
        Return the current number of unison voices.
        """

    @property
    def view(self) -> View:
        """
        Representing the view aspects of a device.
        """

    @property
    def visible_modulation_target_names(self) -> Any:
        """
        Get the names of all the visible modulation targets.
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/ControlSurfaceWrapper.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 12811 bytes
import contextlib
import weakref

import Live
from ableton.v2.base import EventObject, old_hasattr
from ableton.v2.control_surface import MessageScheduler, defaults


def is_real_control_surface(lom_object):
    return is_local_control_surface(lom_object) or isinstance(lom_object, Live.Application.ControlSurfaceProxy)


def is_local_control_surface(lom_object):
    from _Framework import ControlSurface
    from ableton.v2.control_surface import SimpleControlSurface as ControlSurface2
    from ableton.v3.control_surface import ControlSurface as ControlSurface3

    return isinstance(lom_object, (ControlSurface, ControlSurface2, ControlSurface3))


def wrap(control_surface):
    if is_local_control_surface(control_surface):
        return LocalControlSurfaceWrapper(control_surface=control_surface)
    if isinstance(control_surface, Live.Application.ControlSurfaceProxy):
        return RemoteControlSurfaceWrapper(proxy=control_surface)
    return None


class WrapperRegistry:
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._wrapper_registry = {}

    def wrap(self, obj):
        if is_real_control_surface(obj):
            try:
                return self._wrapper_registry[obj]
            except KeyError:
                wrapped = wrap(obj)
                self._wrapper_registry[obj] = wrapped
                with contextlib.suppress(AttributeError):
                    obj.add_disconnect_listener(self._WrapperRegistry__on_control_surface_disconnected)

                return wrapped

            return obj
        return None

    def clear(self):
        for wrapper in self._wrapper_registry.values():
            wrapper.disconnect()

        self._wrapper_registry = {}

    def __on_control_surface_disconnected(self, unwrapped_cs):
        try:
            unwrapped_cs.remove_disconnect_listener(self._WrapperRegistry__on_control_surface_disconnected)
            self._wrapper_registry[unwrapped_cs].disconnect()
            del self._wrapper_registry[unwrapped_cs]
        except KeyError:
            pass


class ControlSurfaceWrapper:
    def disconnect(self):
        raise NotImplementedError

    @property
    def canonical_parent(self):
        pass

    @property
    def type_name(self):
        raise NotImplementedError

    @property
    def control_names(self):
        raise NotImplementedError

    def has_control(self, control):
        raise NotImplementedError

    def get_control_by_name(self, control_name):
        raise NotImplementedError

    def grab_control(self, control):
        raise NotImplementedError

    def release_control(self, control):
        raise NotImplementedError


class LocalControlSurfaceWrapper(ControlSurfaceWrapper):
    def __init__(self, control_surface=None, *a, **k):
        (super(ControlSurfaceWrapper, self).__init__)(*a, **k)
        self._control_surface = control_surface
        self._grabbed_controls = []

    @property
    def __doc__(self):
        return self._control_surface.__doc__

    def set_control_element(self, control, grabbed):
        if old_hasattr(control, "release_parameter"):
            control.release_parameter()
        control.reset()

    def disconnect(self):
        for control in self._grabbed_controls:
            with self._control_surface.component_guard():
                control.resource.release(self)

    def __getattr__(self, name):
        return getattr(self._control_surface, name)

    def __setattr__(self, name, value):
        if name not in ("_control_surface", "_grabbed_controls"):
            setattr(self._control_surface, name, value)
        else:
            super(ControlSurfaceWrapper, self).__setattr__(name, value)

    def __eq__(self, other):
        return self._control_surface == other

    def __hash__(self):
        return hash(self._control_surface)

    @property
    def type_name(self):
        return self._control_surface.__class__.__name__

    @property
    def control_names(self):
        return [
            control.name for control in self._control_surface.controls if old_hasattr(control, "name") if control.name
        ]

    def has_control(self, control):
        return control in self._control_surface.controls

    def get_control_by_name(self, control_name):
        for control in self._control_surface.controls:
            if old_hasattr(control, "name"):
                if control.name == control_name:
                    return control
        return None

    def grab_control(self, control):
        if control not in self._grabbed_controls:
            with self._control_surface.component_guard():
                priority = (
                    self._control_surface.mxd_grab_control_priority()
                    if old_hasattr(self._control_surface, "mxd_grab_control_priority")
                    else 1
                )
                control.resource.grab(self, priority=priority)
                if old_hasattr(control, "suppress_script_forwarding"):
                    control.suppress_script_forwarding = False
                self._grabbed_controls.append(control)

    def release_control(self, control):
        if control in self._grabbed_controls:
            with self._control_surface.component_guard():
                self._grabbed_controls.remove(control)
                control.resource.release(self)


class ControlProxyBase(EventObject):
    __events__ = ("value",)


class ControlProxy(ControlProxyBase):
    def __init__(self, name="", id=None, proxy=None, parent=None, *a, **k):
        (super().__init__)(*a, **k)
        self._name = name
        self._id = id
        self._proxy = proxy
        self._parent = parent

    @property
    def canonical_parent(self):
        return self._parent

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    def send_value(self, *a):
        self._proxy.send_value((self._id, a))

    def receive_value(self, value):
        (self.notify_value)(*value)

    def add_value_listener(self, listener):
        original_count = self.value_listener_count()
        super().add_value_listener(listener)
        if original_count == 0:
            if self.value_listener_count() > 0:
                self._proxy.subscribe_to_control(self._id)

    def remove_value_listener(self, listener):
        original_count = self.value_listener_count()
        super().remove_value_listener(listener)
        if self._proxy:
            if original_count > 0:
                if self.value_listener_count() == 0:
                    self._proxy.unsubscribe_from_control(self._id)


class RemoteControlSurfaceWrapper(ControlSurfaceWrapper):
    def __init__(self, proxy=None, *a, **k):
        (super(ControlSurfaceWrapper, self).__init__)(*a, **k)
        self._proxy = proxy
        self._control_proxies = {
            desc.name: ControlProxy(name=(desc.name), id=(desc.id), proxy=(self._proxy), parent=self) for desc in []
        }
        self._control_proxies_by_id = {p.id: p for p in self._control_proxies.values()}
        proxy.add_control_values_arrived_listener(self._RemoteControlSurfaceWrapper__on_control_values_arrived)
        proxy.add_midi_received_listener(self._RemoteControlSurfaceWrapper__on_midi_received)

        class Timer:
            MS_PER_TICK = defaults.TIMER_DELAY * 1000

            def __init__(self):
                self._timer_instance = None

            def start(self, timeout, callback):
                self_ref = weakref.ref(self)

                def callback_wrapper():
                    callback()
                    if self_ref:
                        self_ref().cancel()

                self._timer_instance = Live.Base.Timer(
                    callback=callback_wrapper,
                    interval=(int((timeout + 1) * self.MS_PER_TICK)),
                    start=True,
                )

            def cancel(self):
                with contextlib.suppress(AttributeError):
                    self._timer_instance.stop()

                self._timer_instance = None

            def disconnect(self):
                self.cancel()

        self._timer = Timer()
        self.mxd_midi_scheduler = MessageScheduler(
            send_message_callback=(self._proxy.send_midi),
            timer=(self._timer),
            on_state_changed_callback=(self._on_mxd_midi_scheduler_state_changed),
        )

    @property
    def timer_instance(self):
        return self._timer._timer_instance

    def disconnect(self):
        self._timer.disconnect()
        self._proxy.enable_receive_midi(False)
        self._proxy.remove_control_values_arrived_listener(self._RemoteControlSurfaceWrapper__on_control_values_arrived)
        self._proxy.remove_midi_received_listener(self._RemoteControlSurfaceWrapper__on_midi_received)

    def __eq__(self, other):
        return self._proxy == other

    def __hash__(self):
        return hash(self._proxy)

    @property
    def type_name(self):
        return self._proxy.type_name

    @property
    def control_names(self):
        return tuple(c.name for c in self._proxy.control_descriptions)

    def _on_mxd_midi_scheduler_state_changed(self, new_state):
        self._proxy.enable_receive_midi(new_state in ("grabbed", "wait", "grabbed_wait"))

    def __on_control_values_arrived(self):
        for control_id, value in self._proxy.fetch_received_values():
            with contextlib.suppress(KeyError):
                self._control_proxies_by_id[control_id].receive_value(value)

    def __on_midi_received(self):
        for message in self._proxy.fetch_received_midi_messages():
            self.mxd_midi_scheduler.handle_message(message)

    def has_control(self, control):
        return control in self._control_proxies.values()

    def get_control_by_name(self, control_name):
        return self._control_proxies.get(control_name)

    def grab_control(self, control):
        if control.id in self._control_proxies_by_id:
            self._proxy.grab_control(control.id)

    def release_control(self, control):
        if control.id in self._control_proxies_by_id:
            self._proxy.release_control(control.id)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/target_track.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7813 bytes
from typing import cast

from ...base import MultiSlot, listenable_property, listens, listens_group
from ...live import liveobj_changed, liveobj_valid, playing_clip_slot, scene_index
from .. import Component
from ..controls import ToggleButtonControl
from ..display import Renderable


class TargetTrackComponent(Component, Renderable):
    lock_button = ToggleButtonControl(color="TargetTrack.LockOff", on_color="TargetTrack.LockOn")

    def __init__(self, name="Target_Track", is_private=False, *a, **k):
        (super().__init__)(a, name=name, is_private=is_private, **k)
        self._target_track = None
        self._target_clip = None
        self._locked_to_track = False
        self.lock_button.connect_property(self, "is_locked_to_track")
        self.register_slot(self.song.view, self._selected_track_changed, "selected_track")
        for event_name in ("arrangement_clips", "playing_slot_index"):
            self.register_slot(
                MultiSlot(
                    subject=self,
                    listener=(self._update_target_clip),
                    event_name_list=("target_track", event_name),
                ),
            )

        self.register_slot(self.application.view, self._update_target_clip, "focused_document_view")
        self.register_slot(self.song.view, self._update_target_clip, "detail_clip")
        self.register_slot(self.song.view, self._update_target_clip, "selected_scene")
        self._selected_track_changed()

    def disconnect(self):
        self._target_track = None
        self._target_clip = None
        super().disconnect()

    @listenable_property
    def target_track(self):
        return self._target_track

    @listenable_property
    def target_clip(self):
        return self._target_clip

    @listenable_property
    def is_locked_to_track(self):
        return self._locked_to_track

    @is_locked_to_track.setter
    def is_locked_to_track(self, is_locked):
        if is_locked != self._locked_to_track:
            self._locked_to_track = is_locked
            current_track = self._target_track
            self._set_target_track()
            self.notify_is_locked_to_track()
            self.notify(
                self.notifications.Track.lock,
                cast(str, self._target_track.name if self._locked_to_track else current_track.name),
                self._locked_to_track,
            )

    def _selected_track_changed(self):
        self._set_target_track()

    def _set_target_track(self):
        if not (self._locked_to_track and liveobj_valid(self._target_track)):
            new_target = self._get_new_target_track()
            if liveobj_changed(self._target_track, new_target):
                self._target_track = new_target
                self._update_target_clip()
                self.notify_target_track()
            self.is_locked_to_track = False

    def _get_new_target_track(self):
        return self.song.view.selected_track

    def _update_target_clip(self):
        self._TargetTrackComponent__on_target_clip_slot_has_clip_changed.subject = None
        clip = None
        if self._target_track in self.song.tracks:
            if self.application.view.focused_document_view == "Session":
                clip = self._target_clip_from_session()
            else:
                clip = self._target_clip_from_arrangement()
        if liveobj_changed(self._target_clip, clip):
            self._target_clip = clip
            self.notify_target_clip()

    def _target_clip_from_arrangement(self):
        clip = self.song.view.detail_clip
        if liveobj_valid(clip):
            if clip in self._target_track.arrangement_clips:
                return clip
            return None
        return None

    def _target_clip_from_session(self):
        clip_slot = playing_clip_slot(self._target_track)
        if not liveobj_valid(clip_slot):
            slot_index = scene_index()
            if slot_index < len(self._target_track.clip_slots):
                clip_slot = self._target_track.clip_slots[slot_index]
        self._TargetTrackComponent__on_target_clip_slot_has_clip_changed.subject = clip_slot
        if clip_slot:
            if clip_slot.has_clip:
                return clip_slot.clip
            return None
        return None

    @listens("has_clip")
    def __on_target_clip_slot_has_clip_changed(self):
        self._update_target_clip()


class ArmedTargetTrackComponent(TargetTrackComponent):
    def __init__(self, *a, **k):
        self._armed_track_list = []
        (super().__init__)(*a, **k)
        self._ArmedTargetTrackComponent__on_tracks_changed.subject = self.song
        self._ArmedTargetTrackComponent__on_tracks_changed()

    def disconnect(self):
        self._armed_track_list = None
        super().disconnect()

    def _selected_track_changed(self):
        if not self._armed_track_list:
            self._set_target_track()

    def _get_new_target_track(self):
        if self._armed_track_list:
            return self._armed_track_list[-1]
        return super()._get_new_target_track()

    @listens("visible_tracks")
    def __on_tracks_changed(self):
        tracks = self._tracks()
        self._ArmedTargetTrackComponent__on_arm_changed.replace_subjects(tracks)
        self._ArmedTargetTrackComponent__on_frozen_state_changed.replace_subjects(tracks)
        self._refresh_armed_track_list()

    @listens_group("arm")
    def __on_arm_changed(self, _):
        self._refresh_armed_track_list()

    @listens_group("is_frozen")
    def __on_frozen_state_changed(self, _):
        self._refresh_armed_track_list()

    def _refresh_armed_track_list(self):
        tracks = self._tracks()
        for track in self._armed_track_list:
            if liveobj_valid(track):
                if track.arm:
                    if not track.is_frozen:
                        if track not in tracks:
                            pass
                    self._armed_track_list.remove(track)

        for track in tracks:
            if track.arm:
                if not track.is_frozen:
                    if track not in self._armed_track_list:
                        self._armed_track_list.append(track)

        self._set_target_track()

    def _tracks(self):
        return [t for t in self.song.tracks if liveobj_valid(t) if t.can_be_armed]



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class AutomationState:
    """
    Class that represent an enumeration of values for AutomationState
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    none = 0
    playing = 1
    overridden = 2


class DeviceParameter:
    """
    This class represents a (automatable) parameter within a MIDI or
    Audio DSP-Device.
    """

    def add_automation_state_listener(self, listener: Callable) -> None:
        """
        add_automation_state_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "automation_state" has changed.

            C++ signature :
                void add_automation_state_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def add_state_listener(self, listener: Callable) -> None:
        """
        add_state_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "state" has changed.

            C++ signature :
                void add_state_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def add_value_listener(self, listener: Callable) -> None:
        """
        add_value_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "value" has changed.

            C++ signature :
                void add_value_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def automation_state_has_listener(self, listener: Callable) -> bool:
        """
        automation_state_has_listener( (DeviceParameter)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "automation_state".

            C++ signature :
                bool automation_state_has_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def begin_gesture(self) -> None:
        """
        begin_gesture( (DeviceParameter)arg1) -> None :
            Notify the begin of a modification of the parameter, when a sequence of modifications have to be consider a consistent group -- for Sexample, when recording automation.

            C++ signature :
                void begin_gesture(TPyHandle<ATimeableValue>)
        """

    def end_gesture(self) -> None:
        """
        end_gesture( (DeviceParameter)arg1) -> None :
            Notify the end of a modification of the parameter. See begin_gesture.

            C++ signature :
                void end_gesture(TPyHandle<ATimeableValue>)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (DeviceParameter)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def re_enable_automation(self) -> None:
        """
        re_enable_automation( (DeviceParameter)arg1) -> None :
            Reenable automation for this parameter.

            C++ signature :
                void re_enable_automation(TPyHandle<ATimeableValue>)
        """

    def remove_automation_state_listener(self, listener: Callable) -> None:
        """
        remove_automation_state_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "automation_state".

            C++ signature :
                void remove_automation_state_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def remove_state_listener(self, listener: Callable) -> None:
        """
        remove_state_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "state".

            C++ signature :
                void remove_state_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def remove_value_listener(self, listener: Callable) -> None:
        """
        remove_value_listener( (DeviceParameter)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "value".

            C++ signature :
                void remove_value_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def state_has_listener(self, listener: Callable) -> bool:
        """
        state_has_listener( (DeviceParameter)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "state".

            C++ signature :
                bool state_has_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    def str_for_value(self, arg2: float) -> str:
        """
        str_for_value( (DeviceParameter)arg1, (float)arg2) -> str :
            Return a string representation of the given value. To be used
            for display purposes only.  This value can include characters like 'db' or
            'hz', depending on the type of the parameter.

            C++ signature :
                TString str_for_value(TPyHandle<ATimeableValue>,float)
        """

    def value_has_listener(self, listener: Callable) -> bool:
        """
        value_has_listener( (DeviceParameter)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "value".

            C++ signature :
                bool value_has_listener(TPyHandle<ATimeableValue>,boost::python::api::object)
        """

    @property
    def automation_state(self) -> Any:
        """
        Returns state of type AutomationState.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the device parameter.
        """

    @property
    def default_value(self) -> int | float:
        """
        Return the default value for this parameter.  A Default value is only
        available for non-quantized parameter types (see 'is_quantized').
        """

    @property
    def is_enabled(self) -> bool:
        """
        Returns false if the parameter has been macro mapped or disabled by Max.
        """

    @property
    def is_quantized(self) -> bool:
        """
        Returns True, if this value is a boolean or integer like switch.
        Non quantized values are continues float values.
        """

    @property
    def max(self) -> int | float:
        """
        Returns const access to the upper value of the allowed range for
        this parameter
        """

    @property
    def min(self) -> int | float:
        """
        Returns const access to the lower value of the allowed range for
        this parameter
        """

    @property
    def name(self) -> str:
        """
        Returns const access the name of this parameter, as visible in Lives
        automation choosers.
        """

    @property
    def original_name(self) -> str:
        """
        Returns const access the original name of this parameter, unaffected of
        any renamings.
        """

    @property
    def short_value_items(self) -> Any:
        """
        Return the list of possible values for this parameter. Like value_items, but prefers short value names if available. Raises an error if 'is_quantized' is False.
        """

    @property
    def state(self) -> Any:
        """
        Returns the state of the parameter:
        - enabled - the parameter's value can be changed,
        - irrelevant - the parameter is enabled, but value changes will not take any effect until it gets enabled,
        - disabled - the parameter's value cannot be changed.
        """

    @property
    def value(self) -> int | float:
        """
        Get/Set the current value (as visible in the GUI) this parameter.
        The value must be inside the min/max properties of this device.
        """

    @value.setter
    def value(self, value: int | float) -> None:
        pass

    @property
    def value_items(self) -> Any:
        """
        Return the list of possible values for this parameter. Raises an error if 'is_quantized' is False.
        """


class ParameterState:
    """
    Class that represent an enumeration of values for ParameterState
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    enabled = 0
    irrelevant = 1
    disabled = 2



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/MidiMap.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1888 bytes
import Live

from .ButtonElement import ButtonElement
from .ButtonMatrixElement import ButtonMatrixElement
from .EncoderElement import EncoderElement
from .SliderElement import SliderElement


def make_button(name, channel, number, midi_message_type):
    is_momentary = True
    return ButtonElement((not is_momentary), midi_message_type, channel, number, name=name)


def make_slider(name, channel, number, midi_message_type):
    return SliderElement(midi_message_type, channel, number, name=name)


def make_encoder(name, channel, number, midi_message_type):
    return EncoderElement(midi_message_type, channel, number, (Live.MidiMap.MapMode.absolute), name=name)


class MidiMap(dict):
    def add_button(self, name, channel, number, midi_message_type):
        self[name] = make_button(name, channel, number, midi_message_type)

    def add_matrix(self, name, element_factory, channel, numbers, midi_message_type):
        def one_dimensional_name(base_name, x, _y):
            return "%s[%d]" % (base_name, x)

        def two_dimensional_name(base_name, x, y):
            return "%s[%d,%d]" % (base_name, x, y)

        name_factory = two_dimensional_name if len(numbers) > 1 else one_dimensional_name
        elements = [
            [
                element_factory(name_factory(name, column, row), channel, identifier, midi_message_type)
                for column, identifier in enumerate(identifiers)
            ]
            for row, identifiers in enumerate(numbers)
        ]
        self[name] = ButtonMatrixElement(rows=elements)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/display_line.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3082 bytes
from typing import Generic, TypeVar, Union

from .. import NotifyingControlElement
from ..display import Text, updating_display

DisplayDataType = TypeVar("DisplayDataType")


class DisplayLineElement(NotifyingControlElement, Generic[DisplayDataType]):
    def __init__(
        self,
        display_fn,
        formatting_fn=lambda message: message.as_ascii(),
        default_formatting=Text(),
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self._display_fn = display_fn
        self._formatting_fn = formatting_fn
        self._default_formatting = default_formatting
        self._last_native_content = None

    def display_message(self, text):
        if updating_display:
            if self._last_native_content != text:
                self._last_native_content = text
                if not self.grabbed:
                    self._do_display(text)
        else:
            self._do_display(text)

    def _apply_default_formatting(self, text: Text):
        if text.max_width is None:
            text.max_width = self._default_formatting.max_width
        if text.justification is None:
            text.justification = self._default_formatting.justification

    def _do_display(self, message: Union[(Text, str)]):
        text = message if isinstance(message, Text) else Text(message)
        self._apply_default_formatting(text)
        self._display_fn(self._formatting_fn(text))

    @property
    def grabbed(self):
        return self.resource.stack_size > 1

    def clear_send_cache(self):
        self._last_native_content = None

    def reset(self):
        if self.grabbed:
            self._do_display("")
        else:
            self._redisplay_last_native_content()

    def _redisplay_last_native_content(self):
        if self._last_native_content is not None:
            self._do_display(self._last_native_content)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/undo_redo.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 631 bytes
from .. import Component
from ..control import ButtonControl


class UndoRedoComponent(Component):
    undo_button = ButtonControl()
    redo_button = ButtonControl()

    @undo_button.pressed
    def undo_button(self, button):
        self._undo()

    @redo_button.pressed
    def redo_button(self, button):
        self._redo()

    def _redo(self):
        if self.song.can_redo:
            self.song.redo()

    def _undo(self):
        if self.song.can_undo:
            self.song.undo()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Capabilities.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2670 bytes
from past.builtins import basestring

GENERIC_SCRIPT_KEY = "generic_script"
PORTS_KEY = "ports"
CONTROLLER_ID_KEY = "controller_id"
TYPE_KEY = "surface_type"
FIRMWARE_KEY = "firmware_version"
AUTO_LOAD_KEY = "auto_load"
VENDORID = "vendor_id"
PRODUCTIDS = "product_ids"
MODEL_NAMES = "model_names"
DIRECTIONKEY = "direction"
PORTNAMEKEY = "name"
MACNAMEKEY = "mac_name"
PROPSKEY = "props"
HIDDEN = "hidden"
SYNC = "sync"
SCRIPT = "script"
NOTES_CC = "notes_cc"
REMOTE = "remote"
PLAIN_OLD_MIDI = "plain_old_midi"


def __create_port_dict(direction, port_name, mac_name, props):
    if props:
        for prop in props:
            pass

    capabilities = {DIRECTIONKEY: direction, PORTNAMEKEY: port_name, PROPSKEY: props}
    if mac_name:
        capabilities[MACNAMEKEY] = mac_name
    return capabilities


def inport(port_name="", props=None, mac_name=None):
    if props is None:
        props = []
    return __create_port_dict("in", port_name, mac_name, props)


def outport(port_name="", props=None, mac_name=None):
    if props is None:
        props = []
    return __create_port_dict("out", port_name, mac_name, props)


def controller_id(vendor_id, product_ids, model_name):
    for product_id in product_ids:
        pass

    model_names = [model_name] if isinstance(model_name, basestring) else model_name
    return {VENDORID: vendor_id, PRODUCTIDS: product_ids, MODEL_NAMES: model_names}



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/InputControlElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 17504 bytes
import contextlib
import logging

from . import Task
from .Dependency import depends
from .Disconnectable import Disconnectable
from .NotifyingControlElement import NotifyingControlElement
from .Signal import Signal
from .SubjectSlot import SubjectEvent
from .Util import const, nop

logger = logging.getLogger(__name__)
MIDI_NOTE_TYPE = 0
MIDI_CC_TYPE = 1
MIDI_PB_TYPE = 2
MIDI_SYSEX_TYPE = 3
MIDI_INVALID_TYPE = 4
MIDI_MSG_TYPES = (MIDI_NOTE_TYPE, MIDI_CC_TYPE, MIDI_PB_TYPE, MIDI_SYSEX_TYPE, MIDI_INVALID_TYPE)
MIDI_NOTE_ON_STATUS = 144
MIDI_NOTE_OFF_STATUS = 128
MIDI_CC_STATUS = 176
MIDI_PB_STATUS = 224


class ParameterSlot(Disconnectable):
    _parameter = None
    _control = None

    def __init__(self, parameter=None, control=None, *a, **k):
        (super().__init__)(*a, **k)
        self.parameter = parameter
        self.control = control

    def _get_control(self):
        return self._control

    def _set_control(self, control):
        if control != self._control:
            self.soft_disconnect()
            self._control = control
            self.connect()

    control = property(_get_control, _set_control)

    def _get_parameter(self):
        return self._parameter

    def _set_parameter(self, parameter):
        if parameter != self._parameter:
            self.soft_disconnect()
            self._parameter = parameter
            self.connect()

    parameter = property(_get_parameter, _set_parameter)

    def connect(self):
        if self._control is not None:
            if self._parameter is not None:
                self._control.connect_to(self._parameter)

    def soft_disconnect(self):
        if self._control is not None:
            if self._parameter is not None:
                self._control.release_parameter()

    def disconnect(self):
        self.parameter = None
        self.control = None
        super().disconnect()


class InputSignal(Signal):
    def __init__(self, sender=None, *a, **k):
        (super().__init__)(a, sender=sender, **k)
        self._input_control = sender

    @contextlib.contextmanager
    def _listeners_update(self):
        try:
            control = self._input_control
            old_count = self.count
            old_wants_forwarding = control.script_wants_forwarding()
            yield
        finally:
            diff_count = self.count - old_count
            control._input_signal_listener_count += diff_count
            if old_wants_forwarding != control.script_wants_forwarding():
                self._input_control._request_rebuild()

    def connect(self, *a, **k):
        with self._listeners_update():
            (super().connect)(*a, **k)

    def disconnect(self, *a, **k):
        with self._listeners_update():
            (super().disconnect)(*a, **k)

    def disconnect_all(self, *a, **k):
        with self._listeners_update():
            (super().disconnect_all)(*a, **k)


class InputControlElement(NotifyingControlElement):
    class ProxiedInterface(NotifyingControlElement.ProxiedInterface):
        send_value = nop
        receive_value = nop
        use_default_message = nop
        set_channel = nop
        message_channel = const(None)
        mapped_parameter = nop
        mapping_sensitivity = const(None)
        reset_state = nop

    __subject_events__ = (SubjectEvent(name="value", signal=InputSignal, override=True),)
    _input_signal_listener_count = 0
    num_delayed_messages = 1
    send_depends_on_forwarding = True

    @depends(request_rebuild_midi_map=(const(nop)))
    def __init__(
        self,
        msg_type=None,
        channel=None,
        identifier=None,
        sysex_identifier=None,
        request_rebuild_midi_map=None,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self._request_rebuild = request_rebuild_midi_map
        self._msg_type = msg_type
        self._msg_channel = channel
        self._msg_identifier = identifier
        self._msg_sysex_identifier = sysex_identifier
        self._original_channel = channel
        self._original_identifier = identifier
        self._needs_takeover = True
        self._is_mapped = True
        self._is_being_forwarded = True
        self._delayed_messages = []
        self._force_next_send = False
        self._mapping_feedback_delay = 0
        self._mapping_sensitivity = 1.0
        self._suppress_script_forwarding = False
        self._send_delayed_messages_task = self._tasks.add(Task.run(self._send_delayed_messages))
        self._send_delayed_messages_task.kill()
        self._parameter_to_map_to = None
        self._in_parameter_gesture = False
        self._last_sent_message = None
        self._report_input = False
        self._report_output = False

    def message_type(self):
        return self._msg_type

    def message_channel(self):
        return self._msg_channel

    def original_channel(self):
        return self._original_channel

    def message_identifier(self):
        return self._msg_identifier

    def original_identifier(self):
        return self._original_identifier

    def message_sysex_identifier(self):
        return self._msg_sysex_identifier

    def message_map_mode(self):
        raise NotImplementedError

    def _get_mapping_sensitivity(self):
        return self._mapping_sensitivity

    def _set_mapping_sensitivity(self, sensitivity):
        self._mapping_sensitivity = sensitivity
        self._request_rebuild()

    mapping_sensitivity = property(_get_mapping_sensitivity, _set_mapping_sensitivity)

    def _get_suppress_script_forwarding(self):
        return self._suppress_script_forwarding

    def _set_suppress_script_forwarding(self, value):
        if self._suppress_script_forwarding != value:
            self._suppress_script_forwarding = value
            self._request_rebuild()

    suppress_script_forwarding = property(_get_suppress_script_forwarding, _set_suppress_script_forwarding)

    def force_next_send(self):
        self._force_next_send = True

    def set_channel(self, channel):
        if self._msg_channel != channel:
            self._msg_channel = channel
            self._request_rebuild()

    def set_identifier(self, identifier):
        if self._msg_identifier != identifier:
            self._msg_identifier = identifier
            self._request_rebuild()

    def set_needs_takeover(self, needs_takeover):
        self._needs_takeover = needs_takeover

    def set_feedback_delay(self, delay):
        self._mapping_feedback_delay = delay

    def needs_takeover(self):
        return self._needs_takeover

    def use_default_message(self):
        if (self._msg_channel, self._msg_identifier) != (self._original_channel, self._original_identifier):
            self._msg_channel = self._original_channel
            self._msg_identifier = self._original_identifier
            self._request_rebuild()

    def _mapping_feedback_values(self):
        value_map = ()
        if self._mapping_feedback_delay != 0:
            if self._msg_type != MIDI_PB_TYPE:
                value_map = tuple(range(128))
            else:
                value_pairs = []
                for value in range(16384):
                    value_pairs.append((value >> 7 & 127, value & 127))

                value_map = tuple(value_pairs)
        return value_map

    def install_connections(self, install_translation, install_mapping, install_forwarding):
        self._send_delayed_messages_task.kill()
        self._is_mapped = False
        self._is_being_forwarded = False
        if self._msg_channel != self._original_channel or self._msg_identifier != self._original_identifier:
            install_translation(
                self._msg_type,
                self._original_identifier,
                self._original_channel,
                self._msg_identifier,
                self._msg_channel,
            )
        if self._parameter_to_map_to is not None:
            self._is_mapped = install_mapping(
                self,
                self._parameter_to_map_to,
                self._mapping_feedback_delay,
                self._mapping_feedback_values(),
            )
        if self.script_wants_forwarding():
            self._is_being_forwarded = install_forwarding(self)
            if self._is_being_forwarded:
                if self.send_depends_on_forwarding:
                    self._send_delayed_messages_task.restart()

    def script_wants_forwarding(self):
        return (
            (not self._suppress_script_forwarding) and (self._input_signal_listener_count > 0) or (self._report_input)
        )

    def begin_gesture(self):
        if self._parameter_to_map_to:
            if not self._in_parameter_gesture:
                self._in_parameter_gesture = True
                self._parameter_to_map_to.begin_gesture()

    def end_gesture(self):
        if self._parameter_to_map_to:
            if self._in_parameter_gesture:
                self._in_parameter_gesture = False
                self._parameter_to_map_to.end_gesture()

    def connect_to(self, parameter):
        if self._parameter_to_map_to != parameter:
            if parameter is None:
                self.release_parameter()
            else:
                self._parameter_to_map_to = parameter
                self._request_rebuild()

    def release_parameter(self):
        if self._parameter_to_map_to is not None:
            self.end_gesture()
            self._parameter_to_map_to = None
            self._request_rebuild()

    def mapped_parameter(self):
        return self._parameter_to_map_to

    def _status_byte(self, channel):
        status_byte = channel
        if self._msg_type == MIDI_NOTE_TYPE:
            status_byte += MIDI_NOTE_ON_STATUS
        else:
            if self._msg_type == MIDI_CC_TYPE:
                status_byte += MIDI_CC_STATUS
            else:
                if self._msg_type == MIDI_PB_TYPE:
                    status_byte += MIDI_PB_STATUS
                else:
                    raise NotImplementedError
        return status_byte

    def identifier_bytes(self):
        if self._msg_type == MIDI_PB_TYPE:
            return ((self._status_byte(self._msg_channel),),)
        if self._msg_type == MIDI_SYSEX_TYPE:
            return (self.message_sysex_identifier(),)
        if self._msg_type == MIDI_NOTE_TYPE:
            return (
                (self._status_byte(self._msg_channel), self.message_identifier()),
                (self._status_byte(self._msg_channel) - 16, self.message_identifier()),
            )
        return ((self._status_byte(self._msg_channel), self.message_identifier()),)

    def _send_delayed_messages(self):
        self.clear_send_cache()
        for value, channel in self._delayed_messages:
            self._do_send_value(value, channel=channel)

        self._delayed_messages[:] = []

    def send_value(self, value, force=False, channel=None):
        value = int(value)
        self._verify_value(value)
        if force or self._force_next_send:
            self._do_send_value(value, channel)
        else:
            if (
                not self.send_depends_on_forwarding
                or self._is_being_forwarded
                and self._send_delayed_messages_task.is_running
            ):
                first = 1 - self.num_delayed_messages
                self._delayed_messages = self._delayed_messages[first:] + [(value, channel)]
            else:
                if (value, channel) != self._last_sent_message:
                    self._do_send_value(value, channel)
        self._force_next_send = False

    def _do_send_value(self, value, channel=None):
        data_byte1 = self._original_identifier
        data_byte2 = value
        status_byte = self._status_byte(self._original_channel if channel is None else channel)
        if self._msg_type == MIDI_PB_TYPE:
            data_byte1 = value & 127
            data_byte2 = value >> 7 & 127
        if self.send_midi((status_byte, data_byte1, data_byte2)):
            self._last_sent_message = (value, channel)
            if self._report_output:
                is_input = True
                self._report_value(value, not is_input)

    def clear_send_cache(self):
        self._last_sent_message = None

    def reset(self):
        self.send_value(0)

    def reset_state(self):
        self.use_default_message()
        self.suppress_script_forwarding = False
        self.release_parameter()

    def receive_value(self, value):
        value = getattr(value, "midi_value", value)
        self._verify_value(value)
        self._last_sent_message = None
        self.notify_value(value)
        if self._report_input:
            is_input = True
            self._report_value(value, is_input)

    def set_report_values(self, report_input, report_output):
        self._report_input = report_input
        self._report_output = report_output

    def _verify_value(self, value):
        if self._msg_type < MIDI_SYSEX_TYPE:
            pass

    def _report_value(self, value, is_input):
        self._verify_value(value)
        message = str(self.__class__.__name__) + " ("
        if self._msg_type == MIDI_NOTE_TYPE:
            message += "Note " + str(self._msg_identifier) + ", "
        else:
            if self._msg_type == MIDI_CC_TYPE:
                message += "CC " + str(self._msg_identifier) + ", "
            else:
                message += "PB "
        message += "Chan. " + str(self._msg_channel)
        message += ") "
        message += "received value " if is_input else "sent value "
        message += str(value)
        logger.debug(message)

    @property
    def _last_sent_value(self):
        if self._last_sent_message:
            return self._last_sent_message[0]
        return -1



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/live/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1786 bytes
from ableton.v2.base import is_parameter_bipolar, liveobj_changed, liveobj_valid, move_current_song_time

from .util import (
    all_visible_tracks,
    any_track_armed,
    application,
    deduplicate_parameters,
    display_name,
    find_parent_track,
    get_bar_length,
    get_parameter_by_name,
    is_arrangement_view_active,
    is_clip_new_recording,
    is_device_rack,
    is_parameter_quantized,
    is_song_recording,
    is_track_armed,
    is_track_recording,
    liveobj_color_to_midi_rgb_values,
    liveobj_color_to_value_from_palette,
    liveobj_name,
    major_version,
    normalized_parameter_value,
    parameter_owner,
    parameter_value_to_midi_value,
    playing_clip_slot,
    prepare_new_clip_slot,
    scene_index,
    set_song,
    song,
    track_index,
)

__all__ = (
    "all_visible_tracks",
    "any_track_armed",
    "application",
    "deduplicate_parameters",
    "display_name",
    "find_parent_track",
    "get_bar_length",
    "get_parameter_by_name",
    "is_arrangement_view_active",
    "is_clip_new_recording",
    "is_device_rack",
    "is_parameter_bipolar",
    "is_parameter_quantized",
    "is_song_recording",
    "is_track_armed",
    "is_track_recording",
    "liveobj_changed",
    "liveobj_color_to_midi_rgb_values",
    "liveobj_color_to_value_from_palette",
    "liveobj_name",
    "liveobj_valid",
    "major_version",
    "move_current_song_time",
    "normalized_parameter_value",
    "parameter_owner",
    "parameter_value_to_midi_value",
    "playing_clip_slot",
    "prepare_new_clip_slot",
    "scene_index",
    "set_song",
    "song",
    "track_index",
)



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class Sample:
    """
    This class represents a sample file loaded into a Simpler instance.
    """

    def add_beats_granulation_resolution_listener(self, listener: Callable) -> None:
        """
        add_beats_granulation_resolution_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "beats_granulation_resolution" has changed.

            C++ signature :
                void add_beats_granulation_resolution_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_beats_transient_envelope_listener(self, listener: Callable) -> None:
        """
        add_beats_transient_envelope_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "beats_transient_envelope" has changed.

            C++ signature :
                void add_beats_transient_envelope_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_beats_transient_loop_mode_listener(self, listener: Callable) -> None:
        """
        add_beats_transient_loop_mode_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "beats_transient_loop_mode" has changed.

            C++ signature :
                void add_beats_transient_loop_mode_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_complex_pro_envelope_listener(self, listener: Callable) -> None:
        """
        add_complex_pro_envelope_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "complex_pro_envelope" has changed.

            C++ signature :
                void add_complex_pro_envelope_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_complex_pro_formants_listener(self, listener: Callable) -> None:
        """
        add_complex_pro_formants_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "complex_pro_formants" has changed.

            C++ signature :
                void add_complex_pro_formants_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_end_marker_listener(self, listener: Callable) -> None:
        """
        add_end_marker_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "end_marker" has changed.

            C++ signature :
                void add_end_marker_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_file_path_listener(self, listener: Callable) -> None:
        """
        add_file_path_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "file_path" has changed.

            C++ signature :
                void add_file_path_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_gain_listener(self, listener: Callable) -> None:
        """
        add_gain_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "gain" has changed.

            C++ signature :
                void add_gain_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_slices_listener(self, listener: Callable) -> None:
        """
        add_slices_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "slices" has changed.

            C++ signature :
                void add_slices_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_slicing_beat_division_listener(self, listener: Callable) -> None:
        """
        add_slicing_beat_division_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "slicing_beat_division" has changed.

            C++ signature :
                void add_slicing_beat_division_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_slicing_region_count_listener(self, listener: Callable) -> None:
        """
        add_slicing_region_count_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "slicing_region_count" has changed.

            C++ signature :
                void add_slicing_region_count_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_slicing_sensitivity_listener(self, listener: Callable) -> None:
        """
        add_slicing_sensitivity_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "slicing_sensitivity" has changed.

            C++ signature :
                void add_slicing_sensitivity_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_slicing_style_listener(self, listener: Callable) -> None:
        """
        add_slicing_style_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "slicing_style" has changed.

            C++ signature :
                void add_slicing_style_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_start_marker_listener(self, listener: Callable) -> None:
        """
        add_start_marker_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "start_marker" has changed.

            C++ signature :
                void add_start_marker_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_texture_flux_listener(self, listener: Callable) -> None:
        """
        add_texture_flux_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "texture_flux" has changed.

            C++ signature :
                void add_texture_flux_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_texture_grain_size_listener(self, listener: Callable) -> None:
        """
        add_texture_grain_size_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "texture_grain_size" has changed.

            C++ signature :
                void add_texture_grain_size_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_tones_grain_size_listener(self, listener: Callable) -> None:
        """
        add_tones_grain_size_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "tones_grain_size" has changed.

            C++ signature :
                void add_tones_grain_size_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_warp_markers_listener(self, listener: Callable) -> None:
        """
        add_warp_markers_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "warp_markers" has changed.

            C++ signature :
                void add_warp_markers_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_warp_mode_listener(self, listener: Callable) -> None:
        """
        add_warp_mode_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "warp_mode" has changed.

            C++ signature :
                void add_warp_mode_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def add_warping_listener(self, listener: Callable) -> None:
        """
        add_warping_listener( (Sample)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "warping" has changed.

            C++ signature :
                void add_warping_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def beat_to_sample_time(self, beat_time: float) -> float:
        """
        beat_to_sample_time( (Sample)self, (float)beat_time) -> float :
            Converts the given beat time to sample time. Raises an error if the sample is not warped.

            C++ signature :
                double beat_to_sample_time(TPyHandle<AMultiSamplePart>,double)
        """

    def beats_granulation_resolution_has_listener(self, listener: Callable) -> bool:
        """
        beats_granulation_resolution_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "beats_granulation_resolution".

            C++ signature :
                bool beats_granulation_resolution_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def beats_transient_envelope_has_listener(self, listener: Callable) -> bool:
        """
        beats_transient_envelope_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "beats_transient_envelope".

            C++ signature :
                bool beats_transient_envelope_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def beats_transient_loop_mode_has_listener(self, listener: Callable) -> bool:
        """
        beats_transient_loop_mode_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "beats_transient_loop_mode".

            C++ signature :
                bool beats_transient_loop_mode_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def clear_slices(self) -> None:
        """
        clear_slices( (Sample)self) -> None :
            Clears all slices created in Simpler's manual mode.

            C++ signature :
                void clear_slices(TPyHandle<AMultiSamplePart>)
        """

    def complex_pro_envelope_has_listener(self, listener: Callable) -> bool:
        """
        complex_pro_envelope_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "complex_pro_envelope".

            C++ signature :
                bool complex_pro_envelope_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def complex_pro_formants_has_listener(self, listener: Callable) -> bool:
        """
        complex_pro_formants_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "complex_pro_formants".

            C++ signature :
                bool complex_pro_formants_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def end_marker_has_listener(self, listener: Callable) -> bool:
        """
        end_marker_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "end_marker".

            C++ signature :
                bool end_marker_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def file_path_has_listener(self, listener: Callable) -> bool:
        """
        file_path_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "file_path".

            C++ signature :
                bool file_path_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def gain_display_string(self) -> str:
        """
        gain_display_string( (Sample)self) -> str :
            Get the gain's display value as a string.

            C++ signature :
                TString gain_display_string(TPyHandle<AMultiSamplePart>)
        """

    def gain_has_listener(self, listener: Callable) -> bool:
        """
        gain_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "gain".

            C++ signature :
                bool gain_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def insert_slice(self, slice_time: int) -> None:
        """
        insert_slice( (Sample)self, (int)slice_time) -> None :
            Add a slice point at the provided time if there is none.

            C++ signature :
                void insert_slice(TPyHandle<AMultiSamplePart>,int)
        """

    def move_slice(self, old_time: int, new_time: int) -> int:
        """
        move_slice( (Sample)self, (int)old_time, (int)new_time) -> int :
            Move the slice point at the provided time.

            C++ signature :
                int move_slice(TPyHandle<AMultiSamplePart>,int,int)
        """

    def remove_beats_granulation_resolution_listener(self, listener: Callable) -> None:
        """
        remove_beats_granulation_resolution_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "beats_granulation_resolution".

            C++ signature :
                void remove_beats_granulation_resolution_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_beats_transient_envelope_listener(self, listener: Callable) -> None:
        """
        remove_beats_transient_envelope_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "beats_transient_envelope".

            C++ signature :
                void remove_beats_transient_envelope_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_beats_transient_loop_mode_listener(self, listener: Callable) -> None:
        """
        remove_beats_transient_loop_mode_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "beats_transient_loop_mode".

            C++ signature :
                void remove_beats_transient_loop_mode_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_complex_pro_envelope_listener(self, listener: Callable) -> None:
        """
        remove_complex_pro_envelope_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "complex_pro_envelope".

            C++ signature :
                void remove_complex_pro_envelope_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_complex_pro_formants_listener(self, listener: Callable) -> None:
        """
        remove_complex_pro_formants_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "complex_pro_formants".

            C++ signature :
                void remove_complex_pro_formants_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_end_marker_listener(self, listener: Callable) -> None:
        """
        remove_end_marker_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "end_marker".

            C++ signature :
                void remove_end_marker_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_file_path_listener(self, listener: Callable) -> None:
        """
        remove_file_path_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "file_path".

            C++ signature :
                void remove_file_path_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_gain_listener(self, listener: Callable) -> None:
        """
        remove_gain_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "gain".

            C++ signature :
                void remove_gain_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_slice(self, slice_time: int) -> None:
        """
        remove_slice( (Sample)self, (int)slice_time) -> None :
            Remove the slice point at the provided time if there is one.

            C++ signature :
                void remove_slice(TPyHandle<AMultiSamplePart>,int)
        """

    def remove_slices_listener(self, listener: Callable) -> None:
        """
        remove_slices_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "slices".

            C++ signature :
                void remove_slices_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_slicing_beat_division_listener(self, listener: Callable) -> None:
        """
        remove_slicing_beat_division_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "slicing_beat_division".

            C++ signature :
                void remove_slicing_beat_division_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_slicing_region_count_listener(self, listener: Callable) -> None:
        """
        remove_slicing_region_count_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "slicing_region_count".

            C++ signature :
                void remove_slicing_region_count_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_slicing_sensitivity_listener(self, listener: Callable) -> None:
        """
        remove_slicing_sensitivity_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "slicing_sensitivity".

            C++ signature :
                void remove_slicing_sensitivity_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_slicing_style_listener(self, listener: Callable) -> None:
        """
        remove_slicing_style_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "slicing_style".

            C++ signature :
                void remove_slicing_style_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_start_marker_listener(self, listener: Callable) -> None:
        """
        remove_start_marker_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "start_marker".

            C++ signature :
                void remove_start_marker_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_texture_flux_listener(self, listener: Callable) -> None:
        """
        remove_texture_flux_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "texture_flux".

            C++ signature :
                void remove_texture_flux_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_texture_grain_size_listener(self, listener: Callable) -> None:
        """
        remove_texture_grain_size_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "texture_grain_size".

            C++ signature :
                void remove_texture_grain_size_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_tones_grain_size_listener(self, listener: Callable) -> None:
        """
        remove_tones_grain_size_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "tones_grain_size".

            C++ signature :
                void remove_tones_grain_size_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_warp_markers_listener(self, listener: Callable) -> None:
        """
        remove_warp_markers_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "warp_markers".

            C++ signature :
                void remove_warp_markers_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_warp_mode_listener(self, listener: Callable) -> None:
        """
        remove_warp_mode_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "warp_mode".

            C++ signature :
                void remove_warp_mode_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def remove_warping_listener(self, listener: Callable) -> None:
        """
        remove_warping_listener( (Sample)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "warping".

            C++ signature :
                void remove_warping_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def reset_slices(self) -> None:
        """
        reset_slices( (Sample)self) -> None :
            Resets all edited slices to their original positions.

            C++ signature :
                void reset_slices(TPyHandle<AMultiSamplePart>)
        """

    def sample_to_beat_time(self, sample_time: float) -> float:
        """
        sample_to_beat_time( (Sample)self, (float)sample_time) -> float :
            Converts the given sample time to beat time. Raises an error if the sample is not warped.

            C++ signature :
                double sample_to_beat_time(TPyHandle<AMultiSamplePart>,double)
        """

    def slices_has_listener(self, listener: Callable) -> bool:
        """
        slices_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "slices".

            C++ signature :
                bool slices_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def slicing_beat_division_has_listener(self, listener: Callable) -> bool:
        """
        slicing_beat_division_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "slicing_beat_division".

            C++ signature :
                bool slicing_beat_division_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def slicing_region_count_has_listener(self, listener: Callable) -> bool:
        """
        slicing_region_count_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "slicing_region_count".

            C++ signature :
                bool slicing_region_count_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def slicing_sensitivity_has_listener(self, listener: Callable) -> bool:
        """
        slicing_sensitivity_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "slicing_sensitivity".

            C++ signature :
                bool slicing_sensitivity_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def slicing_style_has_listener(self, listener: Callable) -> bool:
        """
        slicing_style_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "slicing_style".

            C++ signature :
                bool slicing_style_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def start_marker_has_listener(self, listener: Callable) -> bool:
        """
        start_marker_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "start_marker".

            C++ signature :
                bool start_marker_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def texture_flux_has_listener(self, listener: Callable) -> bool:
        """
        texture_flux_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "texture_flux".

            C++ signature :
                bool texture_flux_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def texture_grain_size_has_listener(self, listener: Callable) -> bool:
        """
        texture_grain_size_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "texture_grain_size".

            C++ signature :
                bool texture_grain_size_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def tones_grain_size_has_listener(self, listener: Callable) -> bool:
        """
        tones_grain_size_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "tones_grain_size".

            C++ signature :
                bool tones_grain_size_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def warp_markers_has_listener(self, listener: Callable) -> bool:
        """
        warp_markers_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "warp_markers".

            C++ signature :
                bool warp_markers_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def warp_mode_has_listener(self, listener: Callable) -> bool:
        """
        warp_mode_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "warp_mode".

            C++ signature :
                bool warp_mode_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    def warping_has_listener(self, listener: Callable) -> bool:
        """
        warping_has_listener( (Sample)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "warping".

            C++ signature :
                bool warping_has_listener(TPyHandle<AMultiSamplePart>,boost::python::api::object)
        """

    @property
    def beats_granulation_resolution(self) -> Any:
        """
        Access to the Granulation Resolution parameter in Beats Warp Mode.
        """

    @property
    def beats_transient_envelope(self) -> Any:
        """
        Access to the Transient Envelope parameter in Beats Warp Mode.
        """

    @property
    def beats_transient_loop_mode(self) -> Any:
        """
        Access to the Transient Loop Mode parameter in Beats Warp Mode.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Access to the sample's canonical parent.
        """

    @property
    def complex_pro_envelope(self) -> Any:
        """
        Access to the Envelope parameter in Complex Pro Mode.
        """

    @property
    def complex_pro_formants(self) -> Any:
        """
        Access to the Formants parameter in Complex Pro Warp Mode.
        """

    @property
    def end_marker(self) -> Any:
        """
        Access to the position of the sample's end marker.
        """

    @property
    def file_path(self) -> Any:
        """
        Get the path of the sample file.
        """

    @property
    def gain(self) -> Any:
        """
        Access to the sample gain.
        """

    @property
    def length(self) -> Any:
        """
        Get the length of the sample file in sample frames.
        """

    @property
    def sample_rate(self) -> Any:
        """
        Access to the audio sample rate of the sample.
        """

    @property
    def slices(self) -> Any:
        """
        Access to the list of slice points in sample time in the sample.
        """

    @property
    def slicing_beat_division(self) -> Any:
        """
        Access to sample's slicing step size.
        """

    @property
    def slicing_region_count(self) -> Any:
        """
        Access to sample's slicing split count.
        """

    @property
    def slicing_sensitivity(self) -> Any:
        """
        Access to sample's slicing sensitivity whose sensitivity is in between 0.0 and 1.0.
        The higher the sensitivity, the more slices will be available.
        """

    @property
    def slicing_style(self) -> Any:
        """
        Access to sample's slicing style.
        """

    @property
    def start_marker(self) -> Any:
        """
        Access to the position of the sample's start marker.
        """

    @property
    def texture_flux(self) -> Any:
        """
        Access to the Flux parameter in Texture Warp Mode.
        """

    @property
    def texture_grain_size(self) -> Any:
        """
        Access to the Grain Size parameter in Texture Warp Mode.
        """

    @property
    def tones_grain_size(self) -> Any:
        """
        Access to the Grain Size parameter in Tones Warp Mode.
        """

    @property
    def warp_markers(self) -> Any:
        """
        Get the warp markers for this sample.
        """

    @property
    def warp_mode(self) -> Any:
        """
        Access to the sample's warp mode.
        """

    @property
    def warping(self) -> Any:
        """
        Access to the sample's warping property.
        """


class SlicingBeatDivision:
    """
    Class that represent an enumeration of values for SlicingBeatDivision
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    sixteenth = 0
    sixteenth_triplett = 1
    eighth = 2
    eighth_triplett = 3
    quarter = 4
    quarter_triplett = 5
    half = 6
    half_triplett = 7
    one_bar = 8
    two_bars = 9
    four_bars = 10


class SlicingStyle:
    """
    Class that represent an enumeration of values for SlicingStyle
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    transient = 0
    beat = 1
    region = 2
    manual = 3


class TransientLoopMode:
    """
    Class that represent an enumeration of values for TransientLoopMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    off = 0
    forward = 1
    alternate = 2



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ModeSelectorComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5404 bytes
from .ControlSurfaceComponent import ControlSurfaceComponent
from .MomentaryModeObserver import MomentaryModeObserver


class ModeSelectorComponent(ControlSurfaceComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._modes_buttons = []
        self._mode_toggle = None
        self._mode_listeners = []
        self._ModeSelectorComponent__mode_index = -1
        self._modes_observers = {}
        self._modes_heap = []

    def _get_protected_mode_index(self):
        return self._ModeSelectorComponent__mode_index

    def _set_protected_mode_index(self, mode):
        self._ModeSelectorComponent__mode_index = mode
        for listener in self._mode_listeners:
            listener()

    _mode_index = property(_get_protected_mode_index, _set_protected_mode_index)

    def _get_public_mode_index(self):
        return self._ModeSelectorComponent__mode_index

    def _set_public_mode_index(self, mode):
        pass

    mode_index = property(_get_public_mode_index, _set_public_mode_index)

    def disconnect(self):
        self._clean_heap()
        if self._mode_toggle is not None:
            self._mode_toggle.remove_value_listener(self._toggle_value)
            self._mode_toggle = None
        self._modes_buttons = None
        self._mode_listeners = None
        super().disconnect()

    def on_enabled_changed(self):
        self.update()

    def set_mode_toggle(self, button):
        if self._mode_toggle is not None:
            self._mode_toggle.remove_value_listener(self._toggle_value)
        self._mode_toggle = button
        if self._mode_toggle is not None:
            self._mode_toggle.add_value_listener(self._toggle_value)
        self.set_mode(0)

    def set_mode_buttons(self, buttons):
        for button in buttons:
            identify_sender = True
            button.add_value_listener(self._mode_value, identify_sender)
            self._modes_buttons.append(button)

        self.set_mode(0)

    def set_mode(self, mode):
        self._clean_heap()
        self._modes_heap = [(mode, None, None)]
        if self._mode_index != mode:
            self._update_mode()

    def _update_mode(self):
        mode = self._modes_heap[-1][0]
        if self._mode_index != mode:
            self._mode_index = mode
            self.update()

    def _clean_heap(self):
        for _, _, observer in self._modes_heap:
            if observer is not None:
                observer.disconnect()

        self._modes_heap = []

    def number_of_modes(self):
        raise NotImplementedError

    def mode_index_has_listener(self, listener):
        return listener in self._mode_listeners

    def add_mode_index_listener(self, listener):
        self._mode_listeners.append(listener)

    def remove_mode_index_listener(self, listener):
        self._mode_listeners.remove(listener)

    def _mode_value(self, value, sender):
        new_mode = self._modes_buttons.index(sender)
        if sender.is_momentary():
            if value > 0:
                mode_observer = MomentaryModeObserver()
                mode_observer.set_mode_details(new_mode, self._controls_for_mode(new_mode), self._get_public_mode_index)
                self._modes_heap.append((new_mode, sender, mode_observer))
                self._update_mode()
            else:
                if self._modes_heap[-1][1] == sender and not self._modes_heap[-1][2].is_mode_momentary():
                    self.set_mode(new_mode)
                else:
                    for mode, button, observer in self._modes_heap:
                        if button == sender:
                            self._modes_heap.remove((mode, button, observer))
                            break

                    self._update_mode()
        else:
            self.set_mode(new_mode)

    def _toggle_value(self, value):
        if not (value != 0 or self._mode_toggle.is_momentary()):
            self.set_mode((self._mode_index + 1) % self.number_of_modes())

    def _controls_for_mode(self, mode):
        pass

    def _on_timer(self):
        for _, _, mode_observer in self._modes_heap:
            if mode_observer is not None:
                mode_observer.on_timer()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 22318 bytes
from contextlib import contextmanager, suppress
from functools import partial, reduce, wraps
from itertools import chain
from numbers import Number

from ableton.v2.base import old_hasattr
from future import standard_library
from future.moves.itertools import zip_longest
from future.utils import raise_

standard_library.install_aliases()


def clamp(val, minv, maxv):
    return max(minv, min(val, maxv))


def linear(minv, maxv, val):
    return minv + (maxv - minv) * val


def nop(*a, **k):
    if a:
        return a[0]
    return None


def negate(value):
    return not value


def const(value):
    return lambda *a, **k: value


def in_range(value, lower_bound, upper_open_bound):
    if not isinstance(value, Number):
        return False
    return value >= lower_bound and value < upper_open_bound


def sign(value):
    if value >= 0.0:
        return 1.0
    return -1.0


def to_slice(obj):
    if isinstance(obj, slice):
        return obj
    if obj != -1:
        return slice(obj, obj + 1)
    return slice(obj, None)


def slice_size(slice, width):
    return len(list(range(width))[slice])


def maybe(fn):
    return lambda x: fn(x) if x is not None else None


def memoize(function):
    memoized = {}

    @wraps(function)
    def wrapper(*args):
        try:
            ret = memoized[args]
        except KeyError:
            ret = memoized[args] = function(*args)

        return ret

    return wrapper


@memoize
def mixin(*args):
    if len(args) == 1:
        return args[0]
    name = "Mixin_{}".format("_".join(cls.__name__ for cls in args))
    return type(str(name), args, {})


def monkeypatch(target, name=None, override=False, doc=None):
    def patcher(func):
        patchname = func.__name__ if name is None else name
        if not override:
            if old_hasattr(target, patchname):
                raise TypeError(f"Class {target.__name__} already has method {patchname}")
        setattr(target, patchname, func)
        with suppress(AttributeError):
            func.__name__ = str(patchname)

        if doc is not None:
            func.__doc__ = doc
        return func

    return patcher


def monkeypatch_extend(target, name=None):
    def patcher(func):
        newfunc = func
        patchname = func.__name__ if name is None else name
        if old_hasattr(target, patchname):
            oldfunc = getattr(target, patchname)
            if not callable(oldfunc):
                raise TypeError("Can not extend non callable attribute")

            @wraps(oldfunc)
            def extended(*a, **k):
                ret = oldfunc(*a, **k)
                func(*a, **k)
                return ret

            newfunc = extended
        else:
            pass
        setattr(target, patchname, newfunc)
        return func

    return patcher


def instance_decorator(decorator):
    class Decorator:
        def __init__(self, func=nop, *args, **kws):
            self.__name__ = func.__name__
            self.__doc__ = func.__doc__
            self._data_name = "%s_%d_decorated_instance" % (func.__name__, id(self))
            self._func = func
            self._args = args
            self._kws = kws

        def __get__(self, obj, cls=None):
            if obj is None:
                return None
            data_name = self._data_name
            try:
                return obj.__dict__[data_name]
            except KeyError:
                decorated = decorator(obj, self._func, *(self._args), **self._kws)
                obj.__dict__[data_name] = decorated
                return decorated

    return Decorator


def forward_property(member):
    class Descriptor:
        def __init__(self, func_or_name):
            self._property_name = func_or_name.__name__ if callable(func_or_name) else func_or_name

        def __get__(self, obj, cls=None):
            return getattr(getattr(obj, member), self._property_name)

        def __set__(self, obj, value):
            return setattr(getattr(obj, member), self._property_name, value)

    return Descriptor


class lazy_attribute:
    def __init__(self, func, name=None):
        wraps(func)(self)
        self._func = func
        if name:
            self.__name__ = name

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        result = obj.__dict__[self.__name__] = self._func(obj)
        return result


def remove_if(predicate, lst):
    return [elem for elem in lst if not predicate(elem)]


def flatten(list):
    return chain(*list)


def group(lst, n):
    return list(zip_longest(*[lst[i::n] for i in range(n)]))


def find_if(predicate, seq):
    for x in seq:
        if predicate(x):
            return x
    return None


def index_if(predicate, seq):
    idx = 0
    for x in seq:
        if predicate(x):
            return idx
        else:
            idx += 1

    return idx


def union(a, b):
    a = dict(a)
    a.update(b)
    return a


def product(iter_a, iter_b):
    for a in iter_a:
        for b in iter_b:
            yield (a, b)


def next(iter):
    return iter.__next__()


def is_iterable(value):
    try:
        it = iter(value)
        return bool(it)
    except TypeError:
        return False


def recursive_map(fn, element, sequence_type=None):
    if sequence_type is None:
        return recursive_map(fn, element, type(element))
    if isinstance(element, sequence_type):
        return [recursive_map(fn, x, sequence_type) for x in element]
    return fn(element)


def chain_from_iterable(iterables):
    for it in iterables:
        yield from it


def is_matrix(iterable):
    if is_iterable(iterable):
        if len(iterable) > 0:
            return all(is_iterable(x) and len(iterable[0]) == len(x) and len(x) > 0 for x in iterable)
    return False


def first(seq):
    return seq[0]


def second(seq):
    return seq[1]


def third(seq):
    return seq[2]


def compose(*funcs):
    return lambda x: reduce(lambda x, f: f(x), funcs[::-1], x)


def is_contextmanager(value):
    return callable(getattr(value, "__enter__", None)) and callable(getattr(value, "__exit__", None))


def infinite_context_manager(generator):
    make_context_manager = contextmanager(generator)

    class InfiniteContextManager:
        def __enter__(self):
            self._delegate = make_context_manager()
            self._delegate.__enter__()

        def __exit__(self, type, err, trace):
            self._delegate.__exit__(type, err, trace)
            del self._delegate

    return InfiniteContextManager


class BooleanContext:
    default_value = False

    def __init__(self, default_value=None, *a, **k):
        (super().__init__)(*a, **k)
        if default_value is not None:
            self.default_value = default_value
        self._current_value = self.default_value

    def __bool__(self):
        return bool(self._current_value)

    def __call__(self, update_value=None):
        return self.Manager(self, update_value)

    @property
    def value(self):
        return self._current_value

    class Manager:
        def __init__(self, managed=None, update_value=None, *a, **k):
            (super(BooleanContext.Manager, self).__init__)(*a, **k)
            self._managed = managed
            self._update_value = update_value if update_value is not None else not managed.default_value

        def __enter__(self):
            managed = self._managed
            self._old_value = managed._current_value
            managed._current_value = self._update_value
            return self

        def __exit__(self, *a, **k):
            self._managed._current_value = self._old_value


def dict_diff(left, right):
    dummy = object()
    return dict([k_v for k_v in iter(right.items()) if left.get(k_v[0], dummy) != k_v[1]])


class NamedTuple:
    def __init__(self, *others, **k):
        super().__init__()
        for other in others:
            diff = dict_diff(self._eq_dict, other._eq_dict)
            self._eq_dict.update(diff)
            self.__dict__.update(diff)

        self.__dict__.update(k)
        if "_eq_dict" in self.__dict__:
            self._eq_dict.update(k)

    def __setattr__(self, name, value):
        raise AttributeError("Named tuple is constant")

    def __delattr__(self, name):
        raise AttributeError("Named tuple is constant")

    def __getitem__(self, name):
        return self.__dict__[name]

    @lazy_attribute
    def _eq_dict(self):
        def public(objdict):
            return dict([k__ for k__ in iter(objdict.items()) if not k__[0].startswith("_")])

        return reduce(
            lambda a, b: union(b, a),
            [public(c.__dict__) for c in self.__class__.__mro__],
            public(self.__dict__),
        )

    def __eq__(self, other):
        return isinstance(other, NamedTuple) and self._eq_dict == other._eq_dict

    def __hash__(self):
        return hash(id(self))

    def __getstate__(self):
        res = dict(self.__dict__)
        with suppress(KeyError):
            del res["_eq_dict"]

        return res


class Slicer:
    def __init__(self, dimensions=1, extractor=nop, keys=(), *a, **k):
        (super().__init__)(*a, **k)
        self._keys = keys
        self._dimensions = dimensions
        self._extractor = extractor

    def __getitem__(self, key):
        new = key if isinstance(key, tuple) else (key,)
        keys = self._keys + new
        if len(keys) == self._dimensions:
            return (self._extractor)(*keys)
        return Slicer(dimensions=(self._dimensions), extractor=(self._extractor), keys=keys)

    def __call__(self):
        return self


get_slice = Slicer()


def slicer(dimensions):
    def decorator(extractor):
        @wraps(extractor)
        def make_slicer(*a, **k):
            return Slicer(dimensions=dimensions, extractor=partial(extractor, *a, **k))

        return make_slicer

    return decorator


def print_message(*messages):
    print(" ".join(map(str, messages)))


class overlaymap:
    def __init__(self, *maps):
        self._maps = maps

    def __getitem__(self, key):
        for m in self._maps:
            if key in m:
                return m[key]

        raise_(KeyError, key)
        return None

    def keys(self):
        res = set()
        for key in chain_from_iterable(self._maps):
            res.add(key)

        return list(res)

    def values(self):
        return [self[key] for key in list(self.keys())]

    def iteritems(self):
        for key in list(self.keys()):
            yield (key, self[key])


def trace_value(value, msg="Value: "):
    print(msg, value)
    return value


class Bindable:
    _bound_instances = None

    def __get__(self, obj, cls=None):
        import weakref

        if obj is None:
            return self
        if self._bound_instances is None:
            self._bound_instances = weakref.WeakKeyDictionary()
        bound_dict = self._bound_instances.setdefault(obj, weakref.WeakKeyDictionary())
        try:
            bound = bound_dict[self]
        except KeyError:
            bound = self.bind(weakref.proxy(obj))
            bound_dict[self] = bound

        return bound

    def bind(self, bind_to_object):
        raise NotImplementedError



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/parameter_mapping_sensitivities.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4703 bytes
from ..live import is_parameter_quantized, liveobj_valid

DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY = 1.0
DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY = 0.1
FINE_GRAIN_SENSITIVITY_FACTOR = 0.1
QUANTIZED_SENSITIVITY_FACTOR = 8


def get_base_quantized_sensitivity(parameter):
    v_range = parameter.max - parameter.min
    is_on_off_parameter = v_range == 1.0
    if is_on_off_parameter:
        return 1.0
    return v_range / QUANTIZED_SENSITIVITY_FACTOR


def parameter_mapping_sensitivities(
    continuous_parameter_sensitivity=DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY,
    quantized_parameter_sensitivity=DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY,
):
    described_sensitivities = create_sensitivities(quantized_parameter_sensitivity)

    def inner(parameter, device):
        default = continuous_parameter_sensitivity
        is_quantized = False
        if liveobj_valid(parameter):
            is_quantized = is_parameter_quantized(parameter, device)
            try:
                default = described_sensitivities[device.class_name][parameter.name]
            except KeyError:
                if is_quantized:
                    default = get_base_quantized_sensitivity(parameter) * quantized_parameter_sensitivity

            return (default, default if is_quantized else default * FINE_GRAIN_SENSITIVITY_FACTOR)
        return None

    return inner


def create_sensitivities(quantized_parameter_sensitivity):
    q_min = quantized_parameter_sensitivity / 2
    q_default = quantized_parameter_sensitivity
    q_median = quantized_parameter_sensitivity * 3
    q_max = quantized_parameter_sensitivity * 6
    return {
        "BeatRepeat": {"Pitch": q_default},
        "GrainDelay": {"Pitch": q_median},
        "InstrumentImpulse": {
            "1 Transpose": q_median,
            "2 Transpose": q_median,
            "3 Transpose": q_median,
            "4 Transpose": q_median,
            "5 Transpose": q_median,
            "6 Transpose": q_median,
            "7 Transpose": q_median,
            "8 Transpose": q_median,
            "Global Transpose": q_median,
        },
        "LoungeLizard": {"Semitone": q_default},
        "MidiArpeggiator": {"Transp. Dist.": q_median},
        "MidiScale": {"Transpose": q_median},
        "MultiSampler": {"O Coarse": q_median},
        "Operator": {"A Coarse": q_median, "B Coarse": q_median, "C Coarse": q_median, "D Coarse": q_median},
        "OriginalSimpler": {
            "Transpose": q_median,
            "Start": q_median,
            "End": q_median,
            "Nudge": q_median,
            "Slice by": q_median,
            "Playback": q_median,
            "Sensitivity": q_max,
            "Division": q_median,
            "Regions": q_max,
        },
        "Resonator": {
            "I Note": q_median,
            "II Pitch": q_median,
            "III Pitch": q_median,
            "IV Pitch": q_median,
            "V Pitch": q_median,
        },
        "StringStudio": {"Octave": q_min, "Semitone": q_default},
        "UltraAnalog": {"OSC1 Octave": q_min, "OSC1 Semi": q_default, "OSC2 Octave": q_min, "OSC2 Semi": q_default},
    }



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/renderable.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6037 bytes
from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Callable

from notifications.type_decl import NOTIFICATION_EVENT_ID, Notification, NotificationParams

from ...base import CompoundDisconnectable, EventObject, EventObjectMeta, const, depends, lazy_attribute, nop
from .notifications import Notifications
from .state import State
from .type_decl import Event

if TYPE_CHECKING:
    from typing_extensions import Unpack


def collect_properties(cls):
    return (
        name for cls in inspect.getmro(cls) for name, _ in EventObjectMeta.collect_listenable_properties(cls.__dict__)
    )


class Renderable(CompoundDisconnectable):
    control_base_type = type(object)
    include_in_top_level_state = True

    @depends(react=(const(None)), notifications=(const(None)), suppress_notifications=(const(None)))
    def __init__(self, react=None, notifications=None, suppress_notifications=None, *a, **k):
        (super().__init__)(*a, **k)
        self._react = react or nop
        self._suppress_notifications = suppress_notifications or nop
        self.notifications = notifications if notifications is not None else Notifications

    @lazy_attribute
    def renderable_state(self):
        renderable_state = State()
        self._init_state_from_listenable_properties(renderable_state)
        self._init_state_from_controls(renderable_state)
        return renderable_state

    def notify(
        self,
        notification: Notification[Callable[[Unpack[NotificationParams]], str | None]],
        *a: Unpack[NotificationParams],
    ):
        if notification is not None:
            data = notification(*a) if callable(notification) else notification
            if data is not None:
                self.dispatch_event(NOTIFICATION_EVENT_ID, data)

    def suppress_notifications(self):
        self._suppress_notifications()

    def _init_state_from_listenable_properties(self, renderable_state):
        if isinstance(self, EventObject):
            for property_name in collect_properties(self.__class__):
                setattr(renderable_state, property_name, getattr(self, property_name))
                self.register_slot(self, self._create_event_handler(property_name), property_name)

    def _init_state_from_controls(self, renderable_state):
        if hasattr(self, "_control_states"):
            for cls in reversed(inspect.getmro(self.__class__)):
                for name, value in vars(cls).items():
                    if isinstance(value, self.control_base_type):
                        control_state = value._get_state(self)
                        if isinstance(control_state, Renderable):
                            setattr(renderable_state, name, control_state.renderable_state)

    def dispatch_event(self, name: str, value):
        self._react(Event(name=name, origin=self, value=value))

    def _create_event_handler(self, property_name):
        def on_event(*_):
            property_value = getattr(self, property_name)
            setattr(
                self.renderable_state,
                property_name,
                property_value.renderable_state if isinstance(property_value, Renderable) else property_value,
            )
            self._react(Event(name=property_name, origin=self, value=property_value))

        return on_event



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/control_surface_specification.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7946 bytes
from types import SimpleNamespace

from . import DEFAULT_BANK_SIZE, BasicColors, DeviceProvider, midi
from .components import AutoArmComponent, BasicRecordingMethod, SessionRingComponent, TargetTrackComponent
from .consts import LOW_PRIORITY
from .default_bank_definitions import BANK_DEFINITIONS
from .default_skin import default_skin
from .display import DisplaySpecification
from .parameter_mapping_sensitivities import (
    DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY,
    DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY,
)


class ControlSurfaceSpecification(SimpleNamespace):
    elements_type = None
    control_surface_skin = default_skin
    display_specification = DisplaySpecification()
    num_tracks = 8
    num_scenes = 1
    include_returns = False
    include_master = False
    right_align_non_player_tracks = False
    snap_track_offset = False
    include_auto_arming = False
    link_session_ring_to_track_selection = False
    link_session_ring_to_scene_selection = False
    session_ring_component_type = SessionRingComponent
    target_track_component_type = TargetTrackComponent
    auto_arm_component_type = AutoArmComponent
    device_provider_type = DeviceProvider
    recording_method_type = BasicRecordingMethod
    feedback_channels = None
    playing_feedback_velocity = BasicColors.ON.midi_value
    recording_feedback_velocity = BasicColors.ON.midi_value
    background_priority = LOW_PRIORITY
    identity_response_id_bytes = None
    custom_identity_response = None
    identity_request = midi.SYSEX_IDENTITY_REQUEST_MESSAGE
    identity_request_delay = 0.0
    hello_messages = None
    goodbye_messages = None
    send_goodbye_messages_last = True
    create_mappings_function = lambda *x: {}
    component_map = {}
    parameter_bank_definitions = BANK_DEFINITIONS
    parameter_bank_size = DEFAULT_BANK_SIZE
    continuous_parameter_sensitivity = DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY
    quantized_parameter_sensitivity = DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/mixer.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9438 bytes
from future.moves.itertools import zip_longest

from ...base import clamp, listens, liveobj_valid
from ..component import Component
from .channel_strip import ChannelStripComponent, release_control


class TrackAssigner:
    def tracks(self, tracks_provider):
        raise NotImplementedError


class SimpleTrackAssigner(TrackAssigner):
    def tracks(self, tracks_provider):
        tracks = list(tracks_provider.controlled_tracks())
        if len(tracks) < tracks_provider.num_tracks:
            num_empty_track_slots = tracks_provider.num_tracks - len(tracks)
            tracks += [None] * num_empty_track_slots
        return tracks


class RightAlignTracksTrackAssigner(TrackAssigner):
    def __init__(self, song=None, include_master_track=False, *a, **k):
        (super().__init__)(*a, **k)
        self._song = song
        self._include_master_track = include_master_track

    def tracks(self, tracks_provider):
        offset = tracks_provider.track_offset
        tracks = tracks_provider.tracks_to_use()
        tracks_to_right_align = list(self._song.return_tracks) + (
            [self._song.master_track] if self._include_master_track else []
        )
        size = tracks_provider.num_tracks
        num_empty_tracks = max(0, size + offset - len(tracks))
        track_list = size * [None]
        for i in range(size):
            track_index = i + offset
            if len(tracks) > track_index:
                track = tracks[track_index]
                empty_offset = 0 if tracks[track_index] not in tracks_to_right_align else num_empty_tracks
                track_list[i + empty_offset] = track

        return track_list


class MixerComponent(Component):
    def __init__(
        self,
        tracks_provider=None,
        track_assigner=None,
        auto_name=False,
        invert_mute_feedback=False,
        channel_strip_component_type=None,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self._channel_strip_component_type = channel_strip_component_type or ChannelStripComponent
        self._track_assigner = (
            track_assigner if track_assigner is not None else RightAlignTracksTrackAssigner(song=(self.song))
        )
        self._provider = tracks_provider
        self._MixerComponent__on_offset_changed.subject = tracks_provider
        self._send_index = 0
        self._prehear_volume_control = None
        self._crossfader_control = None
        self._send_controls = None
        self._channel_strips = []
        self._offset_can_start_after_tracks = False
        for _ in range(self._provider.num_tracks):
            strip = self._channel_strip_component_type(parent=self)
            self._channel_strips.append(strip)
            if invert_mute_feedback:
                strip.set_invert_mute_feedback(True)

        self._master_strip = self._channel_strip_component_type(parent=self)
        self._master_strip.set_track(self.song.master_track)
        self._selected_strip = self._channel_strip_component_type(parent=self)
        self._MixerComponent__on_selected_track_changed.subject = self.song.view
        self._MixerComponent__on_selected_track_changed()
        self._reassign_tracks()
        if auto_name:
            self._auto_name()
        self._MixerComponent__on_track_list_changed.subject = self.song
        self._MixerComponent__on_return_tracks_changed.subject = self.song
        self._MixerComponent__on_return_tracks_changed()

    def disconnect(self):
        super().disconnect()
        release_control(self._prehear_volume_control)
        release_control(self._crossfader_control)
        self._prehear_volume_control = None
        self._crossfader_control = None

    @property
    def send_index(self):
        return self._send_index

    @send_index.setter
    def send_index(self, index):
        if index is None or 0 <= index < self.num_sends:
            if self._send_index != index:
                self._send_index = index
                self.set_send_controls(self._send_controls)
                self.on_send_index_changed()
        else:
            raise IndexError

    def on_send_index_changed(self):
        pass

    @property
    def num_sends(self):
        return len(self.song.return_tracks)

    def channel_strip(self, index):
        return self._channel_strips[index]

    def master_strip(self):
        return self._master_strip

    def selected_strip(self):
        return self._selected_strip

    def set_prehear_volume_control(self, control):
        release_control(self._prehear_volume_control)
        self._prehear_volume_control = control
        self.update()

    def set_crossfader_control(self, control):
        release_control(self._crossfader_control)
        self._crossfader_control = control
        self.update()

    def set_volume_controls(self, controls):
        for strip, control in zip_longest(self._channel_strips, controls or []):
            strip.set_volume_control(control)

    def set_pan_controls(self, controls):
        for strip, control in zip_longest(self._channel_strips, controls or []):
            strip.set_pan_control(control)

    def set_send_controls(self, controls):
        self._send_controls = controls
        for strip, control in zip_longest(self._channel_strips, controls or []):
            if self._send_index is None:
                strip.set_send_controls(None)
            else:
                strip.set_send_controls((None,) * self._send_index + (control,))

    def set_arm_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_arm_button(button)

    def set_solo_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_solo_button(button)

    def set_mute_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_mute_button(button)

    def set_track_select_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_select_button(button)

    def set_shift_button(self, button):
        for strip in self._channel_strips or []:
            strip.set_shift_button(button)

    @listens("offset")
    def __on_offset_changed(self, *a):
        self._reassign_tracks()

    @listens("visible_tracks")
    def __on_track_list_changed(self):
        self._reassign_tracks()

    @listens("selected_track")
    def __on_selected_track_changed(self):
        self._update_selected_strip()
        self._on_selected_track_changed()

    def _update_selected_strip(self):
        selected_track = self.song.view.selected_track
        if liveobj_valid(self._selected_strip):
            if selected_track != self.song.master_track:
                self._selected_strip.set_track(selected_track)
            else:
                self._selected_strip.set_track(None)

    def _on_selected_track_changed(self):
        pass

    @listens("return_tracks")
    def __on_return_tracks_changed(self):
        self._update_send_index()
        self.on_num_sends_changed()

    def _update_send_index(self):
        num_sends = self.num_sends
        if self._send_index is not None:
            self.send_index = clamp(self._send_index, 0, num_sends - 1) if num_sends > 0 else None
        else:
            self.send_index = 0 if num_sends > 0 else None

    def on_num_sends_changed(self):
        pass

    def update(self):
        super().update()
        master_track = self.song.master_track
        if self.is_enabled():
            if self._prehear_volume_control is not None:
                self._prehear_volume_control.connect_to(master_track.mixer_device.cue_volume)
            if self._crossfader_control is not None:
                self._crossfader_control.connect_to(master_track.mixer_device.crossfader)
        else:
            release_control(self._prehear_volume_control)
            release_control(self._crossfader_control)

    def _reassign_tracks(self):
        tracks = self._track_assigner.tracks(self._provider)
        for track, channel_strip in zip(tracks, self._channel_strips):
            channel_strip.set_track(track)

    def _auto_name(self):
        self.name = "Mixer"
        self.master_strip().name = "Master_Channel_Strip"
        self.selected_strip().name = "Selected_Channel_Strip"
        for index, strip in enumerate(self._channel_strips):
            strip.name = "Channel_Strip_%d" % index



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/roar_decoration.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 1023 bytes
from .decoration import LiveObjectDecorator
from .internal_parameter import EnumWrappingParameter


class RoarDeviceDecorator(LiveObjectDecorator):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._additional_parameters = self._create_parameters()
        self.register_disconnectables(self._additional_parameters)

    @property
    def parameters(self):
        return tuple(self._live_object.parameters) + self._additional_parameters

    def _create_parameters(self):
        return (
            EnumWrappingParameter(
                name="Routing",
                parent=self,
                values_host=(self._live_object),
                index_property_host=self,
                values_property="routing_mode_list",
                index_property="routing_mode_index",
            ),
        )
